/// <summary>
/// VizMol.NET.cs
/// 
/// A very simple, quick and dirty molecule viewer....
///
/// </summary>
/// <remarks>
///      Created by  : V.Ganesh
///      Created on  : 7th Nov 2002
///      Modified on : 30th May 2004, flicker free drawing
///      Last updated: 12th June 2004, simple PDB reader added
/// </remarks>

namespace VizMol.NET 
{
    using System;
    using System.IO;
    using System.Collections;
    
    using System.Drawing;
    using System.Windows.Forms;

    using CSDIC.FileTypes;

    /// <summary> 
    /// The main class holding the frame...
    /// </summary>
    public class VizMol : Form {
    
        private MainMenu menu;
        
        private MenuItem mnuFile, mnuHelp;
        
        private String fileName;
    
        private const String appTitle = "VizMol.NET v1.1 - a quick and dirty molecule viewer.";
        
        private Molecule molecule;
        
        private Matrix3D amat = new Matrix3D(), tmat = new Matrix3D(),
                         mat  = new Matrix3D() ;
        
        private float scalefudge = 1.0f;
        
        private float xfac;
        
        private int prevx, prevy;
               
        private float xmin, xmax, ymin, ymax, zmin, zmax;
        
        private Bitmap offScreenBmp;

        private bool mousePressed;

        protected const int X_START = 30;
        protected const int X_END   = 37;
    
        protected const int Y_START = 38;
        protected const int Y_END   = 45;
    
        protected const int Z_START = 46;
        protected const int Z_END   = 53;
    
        // v 2.2 PDB spec earlier didn't have specific field for
        // storing atomic symbols
        protected const int SYMBOL_START_V1 = 13;
        protected const int SYMBOL_END_V1   = 14;
        // v 2.2 PDB spec onwards this field is 76-78
        protected const int SYMBOL_START_V2 = 76;
        protected const int SYMBOL_END_V2   = 78;
    
        // Some relevent PDB keywords
        protected const String ATOM   = "ATOM";
        protected const String HETATM = "HETATM";
        protected const String REMARK = "REMARK";
    
        // the PDB version string for 2.2
        protected const String PDB_VERSION = "FORMAT V. 2.2";

        // the constructor
        public VizMol(string [] args) {
            // set the title of the form
            Text = appTitle;
        
            // set the size of form
            ClientSize = new Size (640, 480);
        
            InitComponents();

            mousePressed = false;

            // enable double buffering
            this.SetStyle(ControlStyles.DoubleBuffer, true);	
            this.SetStyle(ControlStyles.AllPaintingInWmPaint, true);
	    this.SetStyle(ControlStyles.UserPaint, true);

            // open the file if requested
            if (args.Length == 1) {
               fileName = args[0];
               OpenFile(fileName);
            } // end if
        } // end constructor
    
        // method to initilize all the controls
        private void InitComponents() {
            InitMenuBar();       
            RegisterFileTypes();

            this.BackColor = Color.Black;   
        } // end of method InitComponents()     
        
        // register files types with Windows Explorer
        private void RegisterFileTypes() {
            Action action = new Action("open", AppDomain.CurrentDomain.BaseDirectory 
                                               + "VizMol.NET.exe");
            Association assoc = new Association();

            assoc.Actions.Add(action);
            assoc.FileExtensions.Add(".xyz");
 
            ActionService actionService = new ActionService();
            actionService.registerSystemAssociation(assoc);
        }

        // method to initilize the menu bar
        private void InitMenuBar() {
            menu = new MainMenu();
            
            // the file menu
            mnuFile = menu.MenuItems.Add("&File");
            
            mnuFile.MenuItems.Add(new MenuItem("&Open...",
                new EventHandler(OnFileOpen), Shortcut.CtrlO));
            mnuFile.MenuItems.Add("-");
            mnuFile.MenuItems.Add(new MenuItem("E&xit",
            new EventHandler(OnFileExit)));
            
            // and finally the help menu
            mnuHelp = menu.MenuItems.Add("&Help");
                    
            mnuHelp.MenuItems.Add(new MenuItem("&About",
                new EventHandler(OnHelpAbout)));
                        
            // and finally set the menu     
            Menu = menu; 
        } // end of method InitMenuBar()

        // file open hendler
        private void OnFileOpen(object sender, EventArgs e) {
            OpenFileDialog ofd = new OpenFileDialog ();
    
            ofd.Filter = "XYZ files (*.xyz)|*.xyz|PDB files (*.pdb)|*.pdb";  
        
            if (ofd.ShowDialog () == DialogResult.OK) {
                // get the selected file name
                fileName = ofd.FileName;
            
                if (fileName.Length != 0) {
                    OpenFile(fileName);
                } // end if
            } // end if
        } // end of method OnFileOpen()
    
        private void OpenFile(string fileName) {
            try {
                StreamReader sr = new StreamReader(fileName);
                               
                if (fileName.EndsWith(".xyz")) {             
                   LoadXYZMolecule(sr);
                } else if (fileName.EndsWith(".pdb")) {
                   LoadPDBMolecule(sr);
                } else {
                   throw new Exception("Unsupported file format!");
                } // end if
                        
                sr.Close();
                    
                // build the molecule	
                molecule.generateConnectivity();
            
                // find the bounding box         
                FindBB(molecule.atoms);

                // make the paining process
                Invalidate();
                this.Text = appTitle + " - [" + fileName + "]";
            } catch(Exception ex) {
                MessageBox.Show(String.Format("Unable to load {0} : " +
                                ex.ToString() + " ! ", fileName), "Error",
                                MessageBoxButtons.OK);
                Invalidate();
            } // end try .. catch block
        }

        // exit!
        private void OnFileExit(object sender, EventArgs e) {
            Close(); // close every thing and say by for the time being
        } // end of method OnFileExit()
    
        // for mnuHelp...    
        private void OnHelpAbout(object sender, EventArgs e) {
            MessageBox.Show("A very simple, quick and dirty molecule viewer (c) V.Ganesh 2002-2004."
                            , "About...", MessageBoxButtons.OK);
        } // end of method OnHelpAbout
        
        // method to load the molecule; make the connectivity matrix
        // and prepare it for rendering...
        protected void LoadXYZMolecule(StreamReader sr) 
        {
            string line;
            
            molecule = new Molecule();
            
            // the first line has to be number of atoms
            line = sr.ReadLine();
            
            int noOfAtoms = Int32.Parse(line);
            
            // the second line can be ignored .. it is the title
            sr.ReadLine();
            
            // and then read the atom coordinates
            for(int i=0; i<noOfAtoms; i++)
            {
                line = sr.ReadLine();
                 
                char[] chars = {' ', '\t'};                

                string[] lines = line.Split(chars);                       

                // following to remove extra blank lines (characters)!!
                for(int k=0; k<lines.Length; k++) {
                   if (k+1 == lines.Length) break;
                   
                   if (lines[k] == null || lines[k].Trim().Length == 0) {
                      for(int j=k+1; j<lines.Length; j++) {
                         if (!(lines[j] == null) && lines[j].Trim().Length != 0) {
                            lines[k] = lines[j];
                            lines[j] = ""; 
                            break;
                         } // end if
                      } // end for
                   } // end if                   
                } // end for

                // and add them to the molecule object
                molecule.atoms.Add(new Atom(lines[0], 
                                            Double.Parse(lines[1]), 
                                            Double.Parse(lines[2]), 
                                            Double.Parse(lines[3])));
            } // end for                        
        } // end of method LoadXYZMolecule()
        
        // method to load the molecule; make the connectivity matrix
        // and prepare it for rendering...
        protected void LoadPDBMolecule(StreamReader sr) 
        {
            string line, symbol;
            bool isVersion2 = false;
            
            molecule  = new Molecule();

            // read in the PDB entry
            while (true) {
               	line = sr.ReadLine();
                
                if (line == null) break;

                // check for PDB version
                if (!isVersion2) {
                    if (line.IndexOf(REMARK) == 0) {
                        if (line.IndexOf(PDB_VERSION) > 0) {
                            isVersion2 = true;
                        } // end if
                    } // end if
                } // end if

                if ((line.IndexOf(ATOM) == 0) || (line.IndexOf(HETATM) == 0)) {
                    if (isVersion2) {
                       symbol = line.Substring(SYMBOL_START_V2, 
                                               SYMBOL_END_V2-SYMBOL_START_V2);
                    } else {
                       symbol = line.Substring(SYMBOL_START_V1, 
                                               SYMBOL_END_V1-SYMBOL_START_V1); 
                    } // end if
                                        
                    molecule.atoms.Add(new Atom(symbol.Trim(),
                           Double.Parse(line.Substring(X_START, X_END-X_START)),
                           Double.Parse(line.Substring(Y_START, Y_END-Y_START)),
                           Double.Parse(line.Substring(Z_START, Z_END-Z_START))));
                } else {
                    continue;
                } // end if
            } // end while
        } // end of method LoadPDBMolecule()

        // the overridden OnPaint() method as in MFC!
        // here is where i draw the molecule...
        protected override void OnPaint(PaintEventArgs e)
        {
            // get the device context!
            Graphics clientDC = e.Graphics;

            // get the dc associated with off-screen drawing
            Graphics dc = Graphics.FromImage(offScreenBmp);  
            
            dc.FillRectangle(new SolidBrush(Color.Black), 0, 0, this.Width, this.Height);            
            
            // protect the function
            if ((molecule == null) || (molecule.connMat == null))
                return;                            
            
            // draw the molecule
            int theIndex = 0;

            Atom atom1, atom2;
            
            // do the required transformations:
            ArrayList transformedAtoms = new ArrayList();            
                
            float xw = xmax - xmin;
            float yw = ymax - ymin;
            float zw = zmax - zmin;
                
            if (yw > xw)
                xw = yw;
            if (zw > xw)
                xw = zw;
                
            float f1 = this.Width / xw;
            float f2 = this.Height / xw;
                
            xfac = 0.7f * (f1 < f2 ? f1 : f2) * scalefudge;                

            mat.unit();
            mat.translate(-(xmin + xmax) / 2,
                          -(ymin + ymax) / 2,
                          -(zmin + zmax) / 2);
            mat.mult(amat);
            
            mat.scale(xfac, -xfac, 16 * xfac / this.Width);
            mat.translate(this.Width / 2, this.Height / 2, 8);
                
            mat.transform(molecule.atoms, transformedAtoms);
                        
            int noOfAtoms = molecule.atoms.Count;
            for(int i=0; i<noOfAtoms; i++) 
            {
                atom1 = (Atom) transformedAtoms[i];

                Pen pen1 = new Pen((Color) molecule.colorTable[atom1.symbol.ToUpper()]);
                dc.DrawEllipse(pen1, (float) atom1.x, (float) atom1.y, 1.0f, 1.0f);

                for(int j=0; i>j; j++) 
                {                                               
                    if (molecule.connMat[theIndex]) 
                    {
                        atom2 = (Atom) transformedAtoms[j];

                        Atom mid = new Atom("", 
                                              (atom1.x + atom2.x) / 2,
                                              (atom1.y + atom2.y) / 2,
                                               0); // neglet z!
                        Pen pen2 = new Pen((Color) molecule.colorTable[atom2.symbol.ToUpper()]);
                        
                        dc.DrawLine(pen1, (float) atom1.x, (float) atom1.y, (float) mid.x, (float) mid.y);
                        dc.DrawLine(pen2, (float) atom2.x, (float) atom2.y, (float) mid.x, (float) mid.y);
                    } // end if
                    
                    theIndex++;
                } // end for
            } // end for
            
            // finally render the image
            clientDC.DrawImageUnscaled(offScreenBmp, 0, 0);
        } // end of method OnPaint()

        // Resize method
        protected override void OnResize(EventArgs e) 
        {
           offScreenBmp = new Bitmap(this.Width, this.Height);  

           Invalidate();
        }

        // Find the bounding box of this model 
        void FindBB(ArrayList v) 
        {                         
            float xmin = (float) ((Atom) v[0]).x, xmax = xmin;
            float ymin = (float) ((Atom) v[0]).y, ymax = ymin;
            float zmin = (float) ((Atom) v[0]).z, zmax = zmin;
            
            for (int i=1; i<v.Count; i++) {
                float x = (float) ((Atom) v[i]).x;
                
                if (x < xmin)
                    xmin = x;
                if (x > xmax)
                    xmax = x;
                
                float y = (float) ((Atom) v[i]).y;
                
                if (y < ymin)
                    ymin = y;
                if (y > ymax)
                    ymax = y;
                
                float z = (float) ((Atom) v[i]).z;
                
                if (z < zmin)
                    zmin = z;
                if (z > zmax)
                    zmax = z;
            } // end for
            
            this.xmax = xmax;
            this.xmin = xmin;
            this.ymax = ymax;
            this.ymin = ymin;
            this.zmax = zmax;
            this.zmin = zmin;
        } // end of FindDB()        
        
        // the mouse - down handler! ;)
        protected override void OnMouseDown(MouseEventArgs e)
        {
            base.OnMouseDown(e);
            
            prevx = e.X;
            prevy = e.Y;       

            mousePressed = true;     
        } // end of method OnMouseDown()
      
        // the mouse - up handler! ;)
        protected override void OnMouseUp(MouseEventArgs e)
        {
            base.OnMouseUp(e);
            
            int x = e.X;
            int y = e.Y;
            
            tmat.unit();
            
            float xtheta = (prevy - y) * (360.0f / this.Width);
            float ytheta = (x - prevx) * (360.0f / this.Height);
            
            tmat.xrot(xtheta);
            tmat.yrot(ytheta);
            amat.mult(tmat);            
            
            prevx = x;
            prevy = y;      

            mousePressed = false;     

            Invalidate();
        } // end of method OnMouseUp()

        // the mouse - up handler! ;)
        protected override void OnMouseMove(MouseEventArgs e)
        {
            base.OnMouseMove(e);
            
            if (mousePressed) { 
	       int x = e.X;
               int y = e.Y;
            
               tmat.unit();
            
               float xtheta = (prevy - y) * (360.0f / this.Width);
               float ytheta = (x - prevx) * (360.0f / this.Height);
            
               tmat.xrot(xtheta);
               tmat.yrot(ytheta);
               amat.mult(tmat);           
            
               prevx = x;
               prevy = y;   
  
               Invalidate();
            } // end if              
        } // end of method OnMouseMove       

        // the starting point!
        public static void Main(string[] args) {
            Application.Run(new VizMol(args));
        } // end of method Main()
    
    } // end of class VizMol
    
    /// <summary>
    /// class representing an atom
    /// </summary>
    public class Atom 
    {
        public string symbol;
        
        public double x, y, z;
        
        public Atom(string symbol, double x, double y, double z) 
        {
            this.symbol = symbol;
            
            this.x = x;
            this.y = y;
            this.z = z;
        }
        
        public override string ToString()
        {
            return symbol + " : " + x + ", " + y + ", " + z + " ";
        }
    } // end of class Atom
    
    /// <summary>
    /// class representing an molecule
    /// </summary>
    public class Molecule 
    {
        // array of atoms
        public ArrayList atoms;
    
        // the lower triangular connectivity matrix
        public BitArray connMat;
        
        public Hashtable covalentRadius;
        
        public Hashtable colorTable;
        
        public Molecule() 
        {
            atoms = new ArrayList();
            
            // we consider covalentRadius of only few of the atoms
            covalentRadius = new Hashtable();
            
            covalentRadius["H"]  = 0.23;
            covalentRadius["C"]  = 0.77;
            covalentRadius["N"]  = 0.75;
            covalentRadius["O"]  = 0.73;
            covalentRadius["S"]  = 1.02;
            covalentRadius["P"]  = 1.06;
            covalentRadius["NA"] = 1.54;  // Na
            covalentRadius["CL"] = 0.99;  // Cl
            covalentRadius["CA"] = 1.74;  // Ca
            
            // the color table:
            colorTable = new Hashtable();
            
            colorTable["H"]  = Color.LightGray;
            colorTable["C"]  = Color.DarkGray;
            colorTable["N"]  = Color.Blue;
            colorTable["O"]  = Color.Red;
            colorTable["S"]  = Color.Yellow;
            colorTable["P"]  = Color.Orange;
            colorTable["NA"] = Color.Pink;   // Na
            colorTable["CL"] = Color.Green;  // Cl
            colorTable["CA"] = Color.White;  // Ca            
        }
        
        // method to generate the connectivity matrix
        public void generateConnectivity() 
        {
            // initilise the connectivity matrix
            connMat = new BitArray((atoms.Count * (atoms.Count - 1) / 2));
                        
            int theIndex = 0;
            
            for(int i=0; i<atoms.Count; i++) 
            {
                for(int j=0; i>j; j++) 
                {                
                    if (strongBondExist((Atom) atoms[i], (Atom) atoms[j]))
                    {
                        connMat[theIndex] = true; // yes, there is a bond    
                    } 
                    else 
                    {
                        connMat[theIndex] = false; // no bond    
                    } // end if
                    
                    theIndex++;
                } // end for
            } // end for
        } // end of method generateConnectivity()
    
        // method returns true is there exist a strong bond between the two
        // atoms else fase
        private bool strongBondExist(Atom a1, Atom a2)
        {
            double radiiSum = ((Double) covalentRadius[a1.symbol.ToUpper()])
                              + ((Double) covalentRadius[a2.symbol.ToUpper()]);
                              
            double distance = Math.Sqrt(((a2.x-a1.x) * (a2.x-a1.x))
                                        + ((a2.y-a1.y) * (a2.y-a1.y))
                                        + ((a2.z-a1.z) * (a2.z-a1.z)));

            // taken from cambridge cluster database
            return (((radiiSum - 0.4) < distance) 
                   && (distance < (radiiSum + 0.4)));
        } // end of method strongBondExist()
        
    } // end of class Molecule
    
    
    /// <summary>
    /// A fairly conventional 3D matrix object that can transform sets of
    /// 3D points and perform a variety of manipulations on the transform 
    /// </summary>
    public class Matrix3D {
        public float xx, xy, xz, xo;
        public float yx, yy, yz, yo;
        public float zx, zy, zz, zo;
        
        public const double pi = 3.14159265;
        
        // Create a new unit matrix 
        public Matrix3D () 
        {
            xx = 1.0f;
            yy = 1.0f;
            zz = 1.0f;
        }
        
        // Scale by f in all dimensions 
        public void scale(float f) 
        {
            xx *= f;
            xy *= f;
            xz *= f;
            xo *= f;
            yx *= f;
            yy *= f;
            yz *= f;
            yo *= f;
            zx *= f;
            zy *= f;
            zz *= f;
            zo *= f;
        }
        
        // Scale along each axis independently 
        public void scale(float xf, float yf, float zf) 
        {
            xx *= xf;
            xy *= xf;
            xz *= xf;
            xo *= xf;
            yx *= yf;
            yy *= yf;
            yz *= yf;
            yo *= yf;
            zx *= zf;
            zy *= zf;
            zz *= zf;
            zo *= zf;
        }
        
        // Translate the origin 
        public void translate(float x, float y, float z) 
        {
            xo += x;
            yo += y;
            zo += z;
        }
        
        // rotate theta degrees about the y axis 
        public void yrot(double theta) 
        {
            theta *= (pi / 180);
            double ct = Math.Cos(theta);
            double st = Math.Sin(theta);

            float Nxx = (float) (xx * ct + zx * st);
            float Nxy = (float) (xy * ct + zy * st);
            float Nxz = (float) (xz * ct + zz * st);
            float Nxo = (float) (xo * ct + zo * st);

            float Nzx = (float) (zx * ct - xx * st);
            float Nzy = (float) (zy * ct - xy * st);
            float Nzz = (float) (zz * ct - xz * st);
            float Nzo = (float) (zo * ct - xo * st);

            xo = Nxo;
            xx = Nxx;
            xy = Nxy;
            xz = Nxz;
            zo = Nzo;
            zx = Nzx;
            zy = Nzy;
            zz = Nzz;
        }
        
        // rotate theta degrees about the x axis 
        public void xrot(double theta) 
        {
            theta *= (pi / 180);
            double ct = Math.Cos(theta);
            double st = Math.Sin(theta);

            float Nyx = (float) (yx * ct + zx * st);
            float Nyy = (float) (yy * ct + zy * st);
            float Nyz = (float) (yz * ct + zz * st);
            float Nyo = (float) (yo * ct + zo * st);

            float Nzx = (float) (zx * ct - yx * st);
            float Nzy = (float) (zy * ct - yy * st);
            float Nzz = (float) (zz * ct - yz * st);
            float Nzo = (float) (zo * ct - yo * st);

            yo = Nyo;
            yx = Nyx;
            yy = Nyy;
            yz = Nyz;
            zo = Nzo;
            zx = Nzx;
            zy = Nzy;
            zz = Nzz;
        }
        
        // rotate theta degrees about the z axis 
        public void zrot(double theta) 
        {
            theta *= (pi / 180);
            double ct = Math.Cos(theta);
            double st = Math.Sin(theta);

            float Nyx = (float) (yx * ct + xx * st);
            float Nyy = (float) (yy * ct + xy * st);
            float Nyz = (float) (yz * ct + xz * st);
            float Nyo = (float) (yo * ct + xo * st);

            float Nxx = (float) (xx * ct - yx * st);
            float Nxy = (float) (xy * ct - yy * st);
            float Nxz = (float) (xz * ct - yz * st);
            float Nxo = (float) (xo * ct - yo * st);

            yo = Nyo;
            yx = Nyx;
            yy = Nyy;
            yz = Nyz;
            xo = Nxo;
            xx = Nxx;
            xy = Nxy;
            xz = Nxz;
        }
        
        // Multiply this matrix by a second: M = M*R 
        public void mult(Matrix3D rhs) 
        {
            float lxx = xx * rhs.xx + yx * rhs.xy + zx * rhs.xz;
            float lxy = xy * rhs.xx + yy * rhs.xy + zy * rhs.xz;
            float lxz = xz * rhs.xx + yz * rhs.xy + zz * rhs.xz;
            float lxo = xo * rhs.xx + yo * rhs.xy + zo * rhs.xz + rhs.xo;

            float lyx = xx * rhs.yx + yx * rhs.yy + zx * rhs.yz;
            float lyy = xy * rhs.yx + yy * rhs.yy + zy * rhs.yz;
            float lyz = xz * rhs.yx + yz * rhs.yy + zz * rhs.yz;
            float lyo = xo * rhs.yx + yo * rhs.yy + zo * rhs.yz + rhs.yo;

            float lzx = xx * rhs.zx + yx * rhs.zy + zx * rhs.zz;
            float lzy = xy * rhs.zx + yy * rhs.zy + zy * rhs.zz;
            float lzz = xz * rhs.zx + yz * rhs.zy + zz * rhs.zz;
            float lzo = xo * rhs.zx + yo * rhs.zy + zo * rhs.zz + rhs.zo;

            xx = lxx;
            xy = lxy;
            xz = lxz;
            xo = lxo;
            
            yx = lyx;
            yy = lyy;
            yz = lyz;
            yo = lyo;

            zx = lzx;
            zy = lzy;
            zz = lzz;
            zo = lzo;
        }

        // Reinitialize to the unit matrix 
        public void unit() 
        {
            xo = 0;
            xx = 1;
            xy = 0;
            xz = 0;
            yo = 0;
            yx = 0;
            yy = 1;
            yz = 0;
            zo = 0;
            zx = 0;
            zy = 0;
            zz = 1;
        }
        
        /// <summary>
        /// Transform nvert points from v into tv.  v contains the input
        /// coordinates in floating point.  tv ends up holding the transformed
        /// points as integers; three successive entries per point 
        /// </summary>
        public void transform(ArrayList v, ArrayList tv) 
        {
            float lxx = xx, lxy = xy, lxz = xz, lxo = xo;
            float lyx = yx, lyy = yy, lyz = yz, lyo = yo;
            float lzx = zx, lzy = zy, lzz = zz, lzo = zo;
            
            for (int i=0; i<v.Count; i++) {
                float x = (float) ((Atom) v[i]).x;
                float y = (float) ((Atom) v[i]).y;
                float z = (float) ((Atom) v[i]).z;
                
                tv.Add(new Atom(((Atom) v[i]).symbol,
                                (x * lxx + y * lxy + z * lxz + lxo),
                                (x * lyx + y * lyy + z * lyz + lyo),
                                (x * lzx + y * lzy + z * lzz + lzo)));
            }
        }
        
        public override string ToString() {
            return ("[" + xo + "," + xx + "," + xy + "," + xz + ";"
                    + yo + "," + yx + "," + yy + "," + yz + ";"
                        + zo + "," + zx + "," + zy + "," + zz + "]");
        }
        
    } // end of class Matrix3D
    
} // end namespace VizMol.NET
